/* Support of PKCS#7 data structures
 * Copyright (C) 2005 Christian Hoehn, Silvan Geeser
 * Copyright (C) 2002-2005 Andreas Steffen
 * Hochschule fuer Technik Rapperswil, Switzerland
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * RCSID $Id: pkcs7.c,v 1.4 2005/09/09 14:02:11 as Exp $
 */

#include <stdlib.h>
#include <string.h>

#include <freeswan.h>

#include "constants.h"
#include "defs.h"
#include "asn1.h"
#include "oid.h"
#include "log.h"
#include "x509.h"
#include "certs.h"
#include "pkcs7.h"

/* ASN.1 definition of the PKCS#7 ContentInfo type */

static const asn1Object_t contentInfoObjects[] = {
  { 0, "contentInfo",			ASN1_SEQUENCE,     ASN1_NONE }, /*  0 */
  { 1,   "contentType",			ASN1_OID,          ASN1_BODY }, /*  1 */
  { 1,   "content",			ASN1_CONTEXT_C_0,  ASN1_OPT |
							   ASN1_BODY }, /*  2 */
  { 1,   "end opt",			ASN1_EOC,	   ASN1_END  }  /*  3 */
};

#define PKCS7_INFO_TYPE		1
#define PKCS7_INFO_CONTENT	2
#define PKCS7_INFO_ROOF		4

/* ASN.1 definition of the PKCS#7 SignedData type */

static const asn1Object_t signedDataObjects[] = {
  { 0, "signedData",			  ASN1_SEQUENCE,     ASN1_NONE }, /*  0 */
  { 1,   "version",			  ASN1_INTEGER,      ASN1_BODY }, /*  1 */
  { 1,   "digestAlgorithms",		  ASN1_SET,          ASN1_BODY }, /*  2 */
  { 1,   "contentInfo",			  ASN1_SEQUENCE,     ASN1_OBJ  }, /*  3 */
  { 1,   "certificates",		  ASN1_CONTEXT_C_0,  ASN1_OPT |
							     ASN1_LOOP }, /*  4 */
  { 2,      "certificate",		  ASN1_SEQUENCE,     ASN1_OBJ  }, /*  5 */
  { 1,   "end opt or loop",		  ASN1_EOC,	     ASN1_END  }, /*  6 */
  { 1,   "crls",			  ASN1_CONTEXT_C_1,  ASN1_OPT |
							     ASN1_LOOP }, /*  7 */
  { 2,	    "crl",			  ASN1_SEQUENCE,     ASN1_OBJ  }, /*  8 */
  { 1,   "end opt or loop",		  ASN1_EOC,	     ASN1_END  }, /*  9 */
  { 1,   "signerInfos",			  ASN1_SET,	     ASN1_LOOP }, /* 10 */
  { 2,     "signerInfo",		  ASN1_SEQUENCE,     ASN1_OPT  }, /* 11 */
  { 3,       "version",			  ASN1_INTEGER,	     ASN1_BODY }, /* 12 */
  { 3,       "issuerAndSerialNumber",	  ASN1_SEQUENCE,     ASN1_BODY }, /* 13 */
  { 3,       "digestAlgorithm",		  ASN1_SEQUENCE,     ASN1_OBJ  }, /* 14 */
  { 3,       "authenticatedAttributes",	  ASN1_CONTEXT_C_0,  ASN1_OPT |
							     ASN1_OBJ  }, /* 15 */
  { 3,       "end opt",			  ASN1_EOC,	     ASN1_END  }, /* 16 */
  { 3,       "digestEncryptionAlgorithm", ASN1_SEQUENCE,     ASN1_OBJ  }, /* 17 */
  { 3,       "encryptedDigest",		  ASN1_OCTET_STRING, ASN1_BODY }, /* 18 */
  { 3,       "unauthenticatedAttributes", ASN1_CONTEXT_C_1,  ASN1_OPT  }, /* 19 */
  { 3,       "end opt",			  ASN1_EOC,	     ASN1_END  }, /* 20 */
  { 2,     "end opt",			  ASN1_EOC,	     ASN1_END  }, /* 21 */
  { 1,   "end loop",			  ASN1_EOC,	     ASN1_END  }  /* 22 */
};

#define PKCS7_SIGNED_CONTENT_INFO	 3
#define PKCS7_SIGNED_CERT	 	 5
#define PKCS7_DIGEST_ALGORITHM		14
#define PKCS7_AUTH_ATTRIBUTES		15
#define PKCS7_DIGEST_ENC_ALGORITHM	17
#define PKCS7_ENCRYPTED_DIGEST		18
#define PKCS7_SIGNED_ROOF		23

/*
 * Parse PKCS#7 ContentInfo object
 */
bool
parse_pkcs7_contentInfo(chunk_t blob, contentInfo_t *cInfo)
{
    asn1_ctx_t ctx;
    chunk_t object;
    u_int level;
    int objectID = 0;

    asn1_init(&ctx, blob, 0, FALSE, DBG_RAW);

    while (objectID < PKCS7_INFO_ROOF)
    {
	if (!extract_object(contentInfoObjects, &objectID, &object, &level, &ctx))
	     return FALSE;

	if (objectID == PKCS7_INFO_TYPE)
	{
	    cInfo->type = known_oid(object);
	    if (cInfo->type < OID_PKCS7_DATA
	    ||  cInfo->type > OID_PKCS7_ENCRYPTED_DATA)
	    {
		plog("unknown pkcs7 content type");
		return FALSE;
	    }
	}
	else if (objectID == PKCS7_INFO_CONTENT)
	{
	    cInfo->content = object;
	}
	objectID++;
    }
    return TRUE;
}

/*
 * Parse PKCS#7 wrapped X.509 certificates
 */
bool
parse_pkcs7_cert(chunk_t blob, x509cert_t **cert)
{
    asn1_ctx_t ctx;
    chunk_t object;
    u_int level;
    int objectID = 0;

    contentInfo_t cInfo;

    if (!parse_pkcs7_contentInfo(blob, &cInfo))
	return FALSE;

    if (cInfo.type != OID_PKCS7_SIGNED_DATA)
    {
	plog("pkcs7 content type is not signedData");
	return FALSE;
    }

    asn1_init(&ctx, cInfo.content, 2, FALSE, DBG_RAW);

    while (objectID < PKCS7_SIGNED_ROOF)
   {
	if (!extract_object(signedDataObjects, &objectID, &object, &level, &ctx))
	     return FALSE;

	if (objectID == PKCS7_SIGNED_CERT)
	{
	    chunk_t cert_blob;
	    x509cert_t *newcert = alloc_thing(x509cert_t,
					"pkcs7 wrapped x509cert");

	    clonetochunk(cert_blob, object.ptr, object.len, "pkcs7 cert blob");
	    *newcert = empty_x509cert;

	    if (parse_x509cert(cert_blob, level+1, newcert))
	    {
		newcert->next = *cert;
		*cert = newcert;
	    }
	    else
	    {
		free_x509cert(newcert);
	    }
	}
	objectID++;
    }
    return TRUE;
}
